/* The file contains the definition files for the node class */
#include <ctime>
#include <stdlib.h>
#include "node.h"
#include "struct_h.h"
#include <iostream>
#include "constants.h"
using namespace std;

//constructor
node::node(int window, int datasize, double timeout)
{
   W = window;
   source_hacked = 0;
   source_hsent = 1;
   dest_hacked = 1;
   sent_frame_count = 1;
   nextbuffer = NULL;
   nacked_frame_seq = 0;
   state = Normal;
   nacked_frame_to = 0;
   sent_unacked_framecount = 1;
   frame_data_size = datasize;
   frame_TO = timeout;
}

//Intializes the nextbuffer pointer of the framebuffer class
void node::setnextbuff(framebuffer *buffer)
{
   nextbuffer = buffer;
}

/*****************************************************************
This function is invoked when a data frame arrives at the destination node

Input:  receivedframe -> pointer to the arrived frame
        destn -> pointer to the destination node(to which the frame arrives)
        neweventlist -> pointer to the eventlist
        current_time is the event time

output: None

Note: To make the data traffic bi-directional modify this function to
      accept ACK and NACK frames in addition to data frames.
********************************************************************/

void node::receiveframeatdest(frame *receivedframe, node 
*destn, eventlist * neweventlist, double current_time)
{
  if(receivedframe->status == OK)
  {
   if(receivedframe->type == Data)
   {
       if(receivedframe->dataseq == (dest_hacked+1))
       {  //received next expected frame, creating ack
	  frame * newframe = generateframe(ACK, receivedframe->dataseq);
          event *newevent = new event();

          double t = current_time; //+ offset;
	  newevent->scheduleevent(t, newframe, 1,destn, NULL, nextbuffer, NULL, NULL, NULL);

         //create new PQE node for the event: Frame handed from node to 
         //buffer
          PQE *newPQE = new PQE(t, 1, newevent);
          neweventlist->enqueue(*newPQE);      //insert the newly
                                              //created event in the eventlist
          dest_hacked = receivedframe->dataseq;
 	  
       }
       if(receivedframe->dataseq > (dest_hacked+1)) 
       {  //received out-of-order data frame, create and send NACK

          frame * newframe = generateframe(NACK, (dest_hacked+1));
          event *newevent = new event();

	  double t = current_time; // + offset;
          newevent->scheduleevent(t, newframe, 1,destn, NULL, nextbuffer, 
            NULL, NULL, NULL);

          //create new PQE node for event: FRame handed from node to 
          //buffer
          PQE *newPQE = new PQE(t, 1, newevent);
          neweventlist->enqueue(*newPQE);
	  
       }
     }
   }
   else //arrived frame is damaged, send NACK saying expected 
        //(dest_hacked+1)
   {
	  frame * newframe = generateframe(NACK, (dest_hacked+1));
          event *newevent = new event();

          double t = current_time; // + offset;
          newevent->scheduleevent(t, newframe, 1,destn, NULL, nextbuffer, 
	     NULL, NULL, NULL);
          PQE *newPQE = new PQE(t, 1, newevent);
          neweventlist->enqueue(*newPQE);
         
    }
}

/**********************************************
receiveframeatsource() is called when a ACK or NACK arrived at source node

Input: receivedframe -> pointer to the arrived frame
       source -> pointer to the source node(to which the ACK or NACK arrives)
       neweventlist -> pointer to the eventlist
       current_time is the time of the event:ACK or NACK received at source
       flink is the pointer to the link to which the buffer in the 
                source node will give frames to

Output: None

Note: To make data traffic bi-directional, modify this function to accept 
      data frames in addition to ACK and NACK frames.
******************************************************************/

void node::receiveframeatsource(frame *receivedframe, node *source, 
eventlist * neweventlist, double current_time, link *flink)
{
  double newoffset;
  
  if(receivedframe->status == OK)
  {
   if(receivedframe->type == ACK)
   {
     if(receivedframe->ackseq > source_hacked)
     {
        state = Normal;
        sent_unacked_framecount -= (receivedframe->ackseq - source_hacked);
	source_hacked = receivedframe->ackseq;
        
        //received ack, send that many data frames starting from(source_hsent+1)
        //such that the number of sent and unacknowledged frames is always equal to
        //the window size
        int count = (source->W) - sent_unacked_framecount;
        for(int i=(source_hsent+1); i<=(source_hsent+count); i++)
	{
	   frame *new_frame = generateframe(Data, i);
	  
	   sent_frame_count++;
	   sent_unacked_framecount++;

           //create a node to buffer event for the newly generated frame
           event *newevent1 = new event();
           
           //calculation of offset time for successively generated frames

           double SD_data = (double)(new_frame->tlength)/flink->linespeed;
           double SD_ack = (double)(receivedframe->tlength)/flink->linespeed;
           if(SD_data < SD_ack)
             newoffset = SD_data/W;
           else
             newoffset = SD_ack/W;
           if(offset > newoffset)
             offset = newoffset;
           double t;
           if(i == (source_hsent+1))
             t = current_time;
           else
             t = current_time + offset;
           
           newevent1->scheduleevent(t, new_frame, 1,source, NULL, nextbuffer, NULL, NULL, NULL);
           PQE *newPQE = new PQE(t, 1, newevent1);
           neweventlist->enqueue(*newPQE);
          
	   
	   //creating a TO event for the newly generated frame
	   event * newevent2 = new event();
	   
           double to_time = t + frame_TO;
         
	   newevent2->scheduleevent(to_time, new_frame, 0, source, source, 
	     NULL, NULL, NULL, NULL);
	   PQE *newPQE1 = new PQE(to_time, 0, newevent2);
	   neweventlist->enqueue(*newPQE1);  //Add the newly created event to the eventlist
	   
	   current_time = t;
        }
        source_hsent = source_hsent + count;
        
    }
   }
   else
   {       //received NACK from destination
    if(state == Normal)
    {
     if(receivedframe->ackseq == (source_hacked+1))
     {      //received NACK from destination
        nacked_frame_seq = receivedframe->ackseq;
        
        state = Recovery;    //received a NACK from destination, changing state of source 
                             //node to Recovery
      
        //send again frames from (source_hacked+1) to (source_hacked+W)
	   for(int i=(source_hacked+1); i<=(source_hacked+source->W); i++)
           {
           frame *new_frame = generateframe(Data, i);
           sent_frame_count++;

           //create node to frame event
           event *newevent1 = new event();

           //calculation of offset time for successively generated frames
           double SD_data = (double)(new_frame->tlength)/flink->linespeed;
           double SD_ack = (double)(receivedframe->tlength)/flink->linespeed;
           if(SD_data < SD_ack)
             newoffset = SD_data/W;
           else
             newoffset = SD_ack/W;
           if(offset > newoffset)
             offset = newoffset;

           double t;
           if(i == (source_hacked+1))
             t = current_time;
           else
             t = current_time + offset;
           
           newevent1->scheduleevent(t, new_frame, 1,source, NULL, nextbuffer, NULL, NULL, NULL);
           PQE *newPQE = new PQE(t, 1, newevent1);
           neweventlist->enqueue(*newPQE);
           
           //creating a TO event
	   event * newevent2 = new event();
        
           double to_time = t + frame_TO;
           
           if(i == receivedframe->ackseq)
             nacked_frame_to = to_time;
           
           newevent2->scheduleevent(to_time, new_frame, 0, source, source,
             NULL, NULL, NULL, NULL);
           PQE *newPQE1 = new PQE(to_time, 0, newevent2);
           neweventlist->enqueue(*newPQE1);
           current_time = t;
        }
        //changing source_hsent to source_hacked+W
        source_hsent = source_hacked + W;
        sent_unacked_framecount = W;
     }
   }
  }
 }
}
 
/*******************************************************************
   This function is called when a time out event occurs 
   Input: sequence number of the frame for which the time out occurs
          head -> pointer to the eventlist
          source -> pointer to the source node(in which the Time out event occurred)
          PQEptr->pointer to the PQE node which contains the generated TO event
          current time is the time of the event
          flink is the pointer to the link to which the buffer in the 
		source node gives frames to
   output: None
************************************************************************/
void node::timeout(int seqno, eventlist *head, node *source, PQE *PQEptr, 
double current_time, link *flink)
{
    double newoffset;

    if(source_hacked<seqno)
    { //true TO
       PQEptr->ptr->timeout = true;

      //time out occurred, create frames from (source_hacked+1) to 
      //(source_hacked+W)
       for(int i=(source_hacked+1); i<=(source_hacked+W); i++)
        {
           frame *new_frame = generateframe(Data, i);
           sent_frame_count++;
        
           //creating a node to buffer event
           event *newevent1 = new event();

           //calculation of offset time for successively generated frames

           double SD_data = (double)(new_frame->tlength)/flink->linespeed;
           double SD_ack = (double)(ACKFRAME_SIZE)/flink->linespeed;
           if(SD_data < SD_ack)
             newoffset = SD_data/W;
           else
             newoffset = SD_ack/W;
           if(offset > newoffset)
             offset = newoffset;

           double t;
           if(i == (source_hacked+1))
             t = current_time;
           else
             t = current_time + offset;
           newevent1->scheduleevent(t, new_frame, 1,source, NULL, nextbuffer, NULL, NULL, NULL);
           PQE *newPQE = new PQE(t, 1, newevent1);
           
           //Add the newly created event in the eventlist
           head->enqueue(*newPQE);
          
           //creating a TO event
	   event * newevent2 = new event();
          
           double to_time = t + frame_TO;
          
           newevent2->scheduleevent(to_time, new_frame, 0, source, source,
             NULL, NULL, NULL, NULL);
           PQE *newPQE1 = new PQE(to_time, 0, newevent2);
           head->enqueue(*newPQE1);
           current_time = t;
        }
        //changing source_hsent
        source_hsent = source_hacked+W;
    }
}


/************************************************************
 generateframe() - creates new frames
	Input: type -> Type of the frame to be created
               seqno -> sequence number of the frame to be created
        output: Pointer to the created frame
*************************************************************/
frame * node::generateframe(Frame_Type type, int seqno)
{
    frame *newframe = new frame;
    int dlength;

    switch(type)
    {
	case Data:
           newframe->settype(Data);
           dlength = frame_data_size;
           break;
	case ACK:
	   newframe->settype(ACK);
           dlength = ACKDATA_LENGTH;
	   break;
	case NACK:
	   newframe->settype(NACK);
	   dlength = ACKDATA_LENGTH;
	   break;
    };
    
    newframe->setstatus(OK);
    newframe->setdlength(dlength);
    newframe->sethlength(HEADER_LENGTH);
    int tlength = HEADER_LENGTH + dlength;
    newframe->settlength(tlength);
    if(type == Data)
    {
      newframe->setdataseq(seqno);
      newframe->setackseq(0);
    }
    else
    {
      newframe->setackseq(seqno);
      newframe->setdataseq(0);
    }
    return newframe;
}

